-- init.lua is loaded on startup
-- load many quick shortcuts into _G for less verbosity
print("Loading Lua")
require("module")
require("async")
require("bus")
require("util")
print("Super _G loading complete")
